from itertools import groupby

from natsort import os_sorted  # type: ignore

from src.user_types import EditableText, InodesPaths


def get_editable_text(inodes_paths: InodesPaths) -> EditableText:
    """
    Being given a mapping of inodes (ints) to paths (Path objects), return a textual representation
    of these associations, sorted in natural order and grouped by common parent if needed.

    Args:
        inodes_paths: dict of { inode: path }

    Returns:
        A string in either of these formats (tab-separated):
            1. If all files are siblings:
                inode   filename
                inode   filename
                ...
            2. Otherwise:
                common_ancestor/path/to/common_parent
                inode   filename
                inode   filename
                ...

                common_ancestor/path/to/common_parent
                inode   filename
                inode   filename
                ...

                ...

    """
    # Calculate the number of digits in the largest inode (base 10), or 0 if there are no inodes
    inode_size = len(str(max(inodes_paths.keys(), default=0)))

    # Create a naturally sorted list of triples (parent, name, inode) with null paths filtered out
    parents_names_inodes = os_sorted(  # natural sort by parent, then name, then (useless) inode
        (
            f"\n{path.parent}",  # the prefix '\n' adds an empty line for maximal consistency
            path.name,
            str(inode).zfill(inode_size),
        )
        for (inode, path) in inodes_paths.items()
        if path.name  # null paths (empty or root) would make os_sorted raise a ValueError
    )

    # Group them by common parents
    groups = [
        (
            parent,
            list(children),  # converting the inner generator allows to use it twice (len + loop)
        )
        for (parent, children) in groupby(
            parents_names_inodes,
            key=lambda parent_name_inode: parent_name_inode[0],  # group by parent
        )
    ]

    # Format the result
    if len(groups) == 0:  # empty selection
        result = []
    elif len(groups) == 1:  # all files are siblings: showing their parent is useless
        result = [f"{inode}\t{name}" for (_, name, inode) in groups[0][1]]
    else:  # several groups of siblings: showing their parents may be useful for desambiguation
        result = []
        for (parent, children) in groups:
            result.append(f"{parent}")
            result.extend(f"{inode}\t{name}" for (_, name, inode) in children)
    return EditableText("\n".join(result))



from pathlib import Path
from typing import List

from src.user_types import Inode, InodesPaths


def paths_to_inodes_paths(paths: List[Path]) -> InodesPaths:
    """
    Given a list of paths, return a mapping from inodes to paths.

    Args:
        paths: list of Path objects
    
    Raises:
        FileNotFoundError: if any of the paths does not exist.

    Returns:
        A mapping from inodes to paths.
    """
    result = {}
    missing_paths = []
    for path in paths:
        if path.exists():
            result[Inode(path.stat().st_ino)] = path
        else:
            missing_paths.append(path)
    if missing_paths:
        raise FileNotFoundError(f"The following items are missing: {missing_paths}.")
    else:    
        return result



from pathlib import Path

import context
from src.paths_to_inodes_paths import *

import pytest


def test_existing_paths_to_inodes_paths():
    paths = [  # eat your own dog food
        Path("src/paths_to_inodes_paths.py"),
        Path("test/test_paths_to_inodes_paths.py"),
    ]
    result = paths_to_inodes_paths(paths)
    assert all(isinstance(inode, int) for inode in result.keys())
    assert set(result.values()) == set(paths)


def test_non_existing_paths_to_inodes_paths():
    paths = [
        Path("src/paths_to_inodes_paths.py"),
        Path("missing_file_1.txt"),
        Path("missing_file_2.txt"),
    ]
    with pytest.raises(FileNotFoundError):
        paths_to_inodes_paths(paths)


if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



from base64 import b32encode
from itertools import count
from pathlib import Path
from typing import Generator, Iterable, Optional, Set

from pathvalidate import validate_filename


class FileSystem(set):
    def __init__(self, paths: Optional[Iterable[Path]] = None, platform: Optional[str] = None):
        if paths:  # when some initial paths are provided, the file system is considered as pure
            super().__init__(paths)
            for path in self:
                validate_filename(
                    path.name,
                    platform=platform or "auto",
                )  # validate each filename when working with a pure FileSystem
            self.path_exists = lambda path: path in self
            self.siblings = lambda path: self.children(path.parent)
        else:  # otherwise, the file system is considered as concrete
            super().__init__()
            self.path_exists = lambda path: path.exists()
            self.siblings = lambda path: path.parent.glob("*")

    def update_with_source_paths(self, source_paths: Iterable[Path]) -> None:
        """Check all paths exist in the file system and "close" it with their siblings.

        Args:
            source_paths (Iterable[Path]): the paths to be renamed.

        Raises:
            FileNotFoundError: a source path is absent from the file system.
        """
        result: Set[Path] = set()
        for source_path in source_paths:
            if not self.path_exists(source_path):
                raise FileNotFoundError(source_path)
            result.update(self.siblings(source_path))
        self.update(result)  # should not change a pure file system

    def children(self, path: Path) -> Generator[Path, None, None]:
        for candidate in self:
            if candidate.match(f"{path}/*"):
                yield candidate

    def non_existing_sibling(self, path: Path) -> Path:
        """Create the path of a non-existing sibling of a given path.

        Args:
            path (Path): a source path which is the target of another renaming.

        Returns:
            Path: the path to be temporarily used for an intermediate renaming.

        Notes:
            - A previous version relied on `Path.with_stem`, which requires Python 3.9. In the
            present version, the non existing sibling does not conserve the extension of the
            original file.
            - Truncated SHA-256 and BASE-64 were considered, but the former is overkill, and the
            character set of the latter is not appropriate for filenames.
        """
        digest = b32encode(path.name.encode("utf8")).decode("ascii")[:32]
        for suffix in count():
            new_path = path.with_name(f"{digest}-{suffix}")
            if new_path not in self:
                break
        return new_path

    def rename(self, path, new_path):
        """Rename a path into a new path, and rename recursively its descendants.

        The following preconditions are normally satisfied:

        1. `path` and `new_path` are siblings,
        2. `path` is in the file system,
        3. and `new_path` is not in the file system yet.

        Results:
            - `path` is replaced by `new_path` in the file system.
            - Any descendant of `path` is replaced by the appropriate `path`.

        Notes:
            - The renaming is virtual only. The ultimate goal is to produce a sequence of "safe"
                clauses (or arcs) for an ulterior actual renaming.
                Nevertheless, all the consequences of a renaming (specifically, of a folder) are
                simulated to ensure testability.
            - In a virtual file system, renaming a node before its parent is not mandatory.
        """
        offset = len(str(path)) + 1
        for candidate in list(self):
            if candidate == path or str(candidate).startswith(f"{path}/"):
                self.remove(candidate)
                self.add(new_path / str(candidate)[offset:])



from pathlib import Path

import context
from src.renamings import *
from src.user_types import Arc
from src.goodies import rm_tree

import pytest


def test_rename():
    """All renamings succeed."""
    renamer = Renamer(testing=True)
    base = Path("test") / "happy_path"
    rm_tree(base)
    base.mkdir()
    arcs = [
        Arc(base / "source_0", base / "target_0"),
        Arc(base / "source_1", base / "target_1"),
        Arc(base / "source_2", base / "target_2"),
    ]
    for arc in arcs:
        arc.source.touch()
    renamer.perform_renamings(arcs)
    assert set(base.iterdir()) == set(arc.target for arc in arcs)
    print(logger.get_contents())
    assert logger.get_contents() == "\n".join([
        "INFO:root:3 items to rename.",
        "INFO:root:SOURCE:test/happy_path/source_0\tTARGET:test/happy_path/target_0",
        "INFO:root:SOURCE:test/happy_path/source_1\tTARGET:test/happy_path/target_1",
        "INFO:root:SOURCE:test/happy_path/source_2\tTARGET:test/happy_path/target_2",
        "INFO:root:3 items renamed.",
    ])
    rm_tree(base)


def test_rename_and_undo():
    """All renamings succeed, then are successfully undone."""
    renamer = Renamer(testing=True)
    base = Path("test") / "happy_path"
    rm_tree(base)
    base.mkdir()
    arcs = [
        Arc(base / "source_0", base / "target_0"),
        Arc(base / "source_1", base / "target_1"),
        Arc(base / "source_2", base / "target_2"),
    ]
    for arc in arcs:
        arc.source.touch()
    renamer.perform_renamings(arcs)
    arcs_for_undoing = renamer.get_arcs_for_undoing(logger.get_contents())
    renamer.perform_renamings(arcs_for_undoing)
    assert set(base.iterdir()) == set(arc.source for arc in arcs)
    print(logger.get_contents())
    assert logger.get_contents() == "\n".join([
        "INFO:root:3 items to rename.",
        "INFO:root:SOURCE:test/happy_path/source_0\tTARGET:test/happy_path/target_0",
        "INFO:root:SOURCE:test/happy_path/source_1\tTARGET:test/happy_path/target_1",
        "INFO:root:SOURCE:test/happy_path/source_2\tTARGET:test/happy_path/target_2",
        "INFO:root:3 items renamed.",
        "INFO:root:3 items to rename.",
        "INFO:root:SOURCE:test/happy_path/target_2\tTARGET:test/happy_path/source_2",
        "INFO:root:SOURCE:test/happy_path/target_1\tTARGET:test/happy_path/source_1",
        "INFO:root:SOURCE:test/happy_path/target_0\tTARGET:test/happy_path/source_0",
        "INFO:root:3 items renamed.",
    ])
    rm_tree(base)


def test_rename_and_undo_fail_after_target_deletion():
    """All renamings succeed, but undo them fails."""
    renamer = Renamer(testing=True)
    base = Path("test") / "happy_path"
    rm_tree(base)
    base.mkdir()
    arcs = [
        Arc(base / "source_0", base / "target_0"),
        Arc(base / "source_1", base / "target_1"),
        Arc(base / "source_2", base / "target_2"),
    ]
    for arc in arcs:
        arc.source.touch()
    renamer.perform_renamings(arcs)
    assert set(base.iterdir()) == set(arc.target for arc in arcs)
    arcs[1].target.unlink()  # delete a renamed file
    arcs_for_undoing = renamer.get_arcs_for_undoing(logger.get_contents())
    with pytest.raises(RecoverableRenamingError):
        renamer.perform_renamings(arcs_for_undoing)
    assert set(base.iterdir()) == {
        arcs[2].source, # correctly undone
        arcs[0].target, # failed to be undone
    }
    renamer.rollback_renamings()
    print(set(base.iterdir()))
    assert set(base.iterdir()) == {
        arcs[2].target, # correctly undone
        arcs[0].target, # failed to be undone
    }
    print(logger.get_contents())
    assert logger.get_contents() == "\n".join([
        "INFO:root:3 items to rename.",
        "INFO:root:SOURCE:test/happy_path/source_0\tTARGET:test/happy_path/target_0",
        "INFO:root:SOURCE:test/happy_path/source_1\tTARGET:test/happy_path/target_1",
        "INFO:root:SOURCE:test/happy_path/source_2\tTARGET:test/happy_path/target_2",
        "INFO:root:3 items renamed.",
        "INFO:root:3 items to rename.",
        "INFO:root:SOURCE:test/happy_path/target_2\tTARGET:test/happy_path/source_2",
        "WARNING:root:perform_renamings: [Errno 2] No such file or directory: 'test/happy_path/target_1' -> 'test/happy_path/source_1'",
        "INFO:root:1 renaming to roll back.",
        "INFO:root:SOURCE:test/happy_path/source_2\tTARGET:test/happy_path/target_2",
        "INFO:root:1 renaming rolled back.",
    ])
    rm_tree(base)


def test_rename_and_undo_fail_after_log_file_deletion():
    """All renamings succeed, but undo them fails."""
    renamer = Renamer(testing=True)
    base = Path("test") / "happy_path"
    rm_tree(base)
    base.mkdir()
    arcs = [
        Arc(base / "source_0", base / "target_0"),
        Arc(base / "source_1", base / "target_1"),
        Arc(base / "source_2", base / "target_2"),
    ]
    for arc in arcs:
        arc.source.touch()
    renamer.perform_renamings(arcs)
    assert set(base.iterdir()) == set(arc.target for arc in arcs)
    logger.path.unlink()  # delete the log file
    arcs_for_undoing = renamer.get_arcs_for_undoing(logger.get_contents())
    assert arcs_for_undoing == []
    assert set(base.iterdir()) == set(arc.target for arc in arcs)
    rm_tree(base)


def test_rename_fail_and_rollback():
    """One renaming fails, but the previous ones are successfully rolled back."""
    renamer = Renamer(testing=True)
    base = Path("test") / "rollback_path"
    rm_tree(base)
    base.mkdir()
    arcs = [
        Arc(base / "source_0", base / "target_0"),
        Arc(base / "source_1", base / "target_1"),
        Arc(base / "source_2", base / "target_2"),
        Arc(base / "source_3", base / "target_3"),
    ]
    for arc in arcs:
        arc.source.touch()
    arcs[2].source.unlink()  # delete a source
    with pytest.raises(RecoverableRenamingError):
        renamer.perform_renamings(arcs) # the renaming fails
    renamer.rollback_renamings() # but we can roll back the previous ones
    assert set(base.iterdir()) == {
        base / "source_0", # rolled back
        base / "source_1", # rolled back
        base / "source_3", # untouched
    }
    print(logger.get_contents())
    assert logger.get_contents() == "\n".join([
        "INFO:root:4 items to rename.",
        "INFO:root:SOURCE:test/rollback_path/source_0\tTARGET:test/rollback_path/target_0",
        "INFO:root:SOURCE:test/rollback_path/source_1\tTARGET:test/rollback_path/target_1",
        "WARNING:root:perform_renamings: [Errno 2] No such file or directory: 'test/rollback_path/source_2' -> 'test/rollback_path/target_2'",
        "INFO:root:2 renamings to roll back.",
        "INFO:root:SOURCE:test/rollback_path/target_1\tTARGET:test/rollback_path/source_1",
        "INFO:root:SOURCE:test/rollback_path/target_0\tTARGET:test/rollback_path/source_0",
        "INFO:root:2 renamings rolled back.",
    ])
    rm_tree(base)


def test_rename_fail_and_rollback_and_undo():
    """Undoing a successful rollback is possible, but has no effect."""
    renamer = Renamer(testing=True)
    base = Path("test") / "rollback_path"
    rm_tree(base)
    base.mkdir()
    arcs = [
        Arc(base / "source_0", base / "target_0"),
        Arc(base / "source_1", base / "target_1"),
        Arc(base / "source_2", base / "target_2"),
        Arc(base / "source_3", base / "target_3"),
    ]
    for arc in arcs:
        arc.source.touch()
    arcs[2].source.unlink()  # delete a source
    with pytest.raises(RecoverableRenamingError):
        renamer.perform_renamings(arcs) # the renaming fails
    renamer.rollback_renamings() # but we can roll back the previous ones
    arcs_for_undoing = renamer.get_arcs_for_undoing(logger.get_contents())
    renamer.perform_renamings(arcs_for_undoing)
    assert set(base.iterdir()) == {
        base / "source_0", # rolled back
        base / "source_1", # rolled back
        base / "source_3", # untouched
    }
    print(logger.get_contents())
    assert logger.get_contents() == "\n".join([
        "INFO:root:4 items to rename.",
        "INFO:root:SOURCE:test/rollback_path/source_0\tTARGET:test/rollback_path/target_0",
        "INFO:root:SOURCE:test/rollback_path/source_1\tTARGET:test/rollback_path/target_1",
        "WARNING:root:perform_renamings: [Errno 2] No such file or directory: 'test/rollback_path/source_2' -> 'test/rollback_path/target_2'",
        "INFO:root:2 renamings to roll back.",
        "INFO:root:SOURCE:test/rollback_path/target_1\tTARGET:test/rollback_path/source_1",
        "INFO:root:SOURCE:test/rollback_path/target_0\tTARGET:test/rollback_path/source_0",
        "INFO:root:2 renamings rolled back.",
        "INFO:root:4 items to rename.",
        "INFO:root:SOURCE:test/rollback_path/source_0\tTARGET:test/rollback_path/target_0",
        "INFO:root:SOURCE:test/rollback_path/source_1\tTARGET:test/rollback_path/target_1",
        "INFO:root:SOURCE:test/rollback_path/target_1\tTARGET:test/rollback_path/source_1",
        "INFO:root:SOURCE:test/rollback_path/target_0\tTARGET:test/rollback_path/source_0",
        "INFO:root:4 items renamed.",
    ])
    rm_tree(base)


def test_rename_fail_and_rollback_fail():
    """One renaming fails, and the previous ones fail to be rolled back."""
    renamer = Renamer(testing=True)
    base = Path("test") / "rollback_path"
    rm_tree(base)
    base.mkdir()
    arcs = [
        Arc(base / "source_0", base / "target_0"),
        Arc(base / "source_1", base / "target_1"),
        Arc(base / "source_2", base / "target_2"),
        Arc(base / "source_3", base / "target_3"),
    ]
    for arc in arcs:
        arc.source.touch()
    arcs[2].source.unlink()  # delete a source
    with pytest.raises(RecoverableRenamingError):
        renamer.perform_renamings(arcs) # the renaming fails
    arcs[1].target.unlink() # delete a file already renamed
    with pytest.raises(FileNotFoundError):
       renamer.rollback_renamings() # the rollback fails
    print(logger.get_contents())
    assert logger.get_contents() == "\n".join([
        "INFO:root:4 items to rename.",
        "INFO:root:SOURCE:test/rollback_path/source_0\tTARGET:test/rollback_path/target_0",
        "INFO:root:SOURCE:test/rollback_path/source_1\tTARGET:test/rollback_path/target_1",
        "WARNING:root:perform_renamings: [Errno 2] No such file or directory: 'test/rollback_path/source_2' -> 'test/rollback_path/target_2'",
        "INFO:root:2 renamings to roll back.",
        "ERROR:root:rollback_renamings: [Errno 2] No such file or directory: 'test/rollback_path/target_1' -> 'test/rollback_path/source_1'",
    ])
    with pytest.raises(Exception):
        renamer.undo_renamings() # The previous rollback failed. Undoing is not possible.
    rm_tree(base)


def test_undo_with_empty_log_file():
    renamer = Renamer(testing=True)
    arcs_for_undoing = renamer.get_arcs_for_undoing(logger.get_contents()) # doesn't raise an exception
    assert arcs_for_undoing == []

if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



import pytest

from pathlib import Path

import context
from src.file_system import FileSystem
from pathvalidate import ValidationError


@pytest.fixture(scope="module")
def paths():
    return [Path(line) for line in Path("test/fhs.txt").read_text().strip().split("\n")]


@pytest.fixture()
def fs(paths):
    return FileSystem(paths)


def test_constructor(fs):
    assert Path("/usr/local") in fs


path_and_platform_data = [
    ([Path("/foo/bar\0")], "linux"),
    ([Path("/foo/bar\0")], "macOS"),
    ([Path("/foo/bar?")], "Windows"),
]

@pytest.mark.parametrize("paths, platform", path_and_platform_data)
def test_invalid_pure_filenames(paths, platform):
    with pytest.raises(ValidationError):
        FileSystem(paths, platform)


def test_update_with_source_paths_concrete():
    fs = FileSystem()
    paths = [
        Path("./src/goodies.py"),
        Path("."),
    ]
    fs.update_with_source_paths(paths)
    assert Path("./src/file_system.py") in fs  # sibling of `goodies.py`
    assert Path("./test") in fs  # sibling of `.`
    assert Path("./LICENSE") in fs  # sibling of `.`
    assert Path("./test/test_file_system.py") not in fs  # child of a sibling of `.`


def test_update_with_source_paths_not_existing(fs):
    paths = [
        Path("/foo/bar"),
    ]
    with pytest.raises(FileNotFoundError) as offending_path:
        fs.update_with_source_paths(paths)
    assert offending_path.value.args[0] == Path("/foo/bar")


def test_children(fs):
    expected = {
        "/usr/X11R6",
        "/usr/bin",
        "/usr/etc",
        "/usr/games",
        "/usr/include",
        "/usr/lib",
        "/usr/libexec",
        "/usr/local",
        "/usr/sbin",
        "/usr/share",
        "/usr/src",
        "/usr/tmp",
    }
    result = set(map(str, fs.children(Path("/usr"))))
    assert result == expected


def test_siblings(fs):
    expected = {
        Path("/usr/X11R6"),
        Path("/usr/bin"),
        Path("/usr/etc"),
        Path("/usr/games"),
        Path("/usr/include"),
        Path("/usr/lib"),
        Path("/usr/libexec"),
        Path("/usr/local"),  # a node is considered as its own sibling
        Path("/usr/sbin"),
        Path("/usr/share"),
        Path("/usr/src"),
        Path("/usr/tmp"),
    }
    result = set(fs.siblings(Path("/usr/local")))
    assert result == expected


def test_non_existing_sibling_folder(fs):
    expected = "/usr/NRXWGYLM-0"
    result = str(fs.non_existing_sibling(Path("/usr/local")))
    assert result == expected


def test_non_existing_sibling_file(fs):
    expected = "/etc/PBUW4ZLUMQXGI===-0"
    result = str(fs.non_existing_sibling(Path("/etc/xinetd.d")))
    assert result == expected


def test_non_existing_sibling_file_with_collision(fs):
    fs.add(Path("/etc/PBUW4ZLUMQXGI===-0"))
    expected = "/etc/PBUW4ZLUMQXGI===-1"
    result = str(fs.non_existing_sibling(Path("/etc/xinetd.d")))
    assert result == expected


def test_rename_leaf(fs):
    original_fs = set(fs)
    fs.rename(Path("/mnt/floppy"), Path("/mnt/toaster"))
    assert original_fs - fs == {Path("/mnt/floppy")}
    assert fs - original_fs == {Path("/mnt/toaster")}


def test_rename_internal_node(fs):
    original_fs = set(fs)
    fs.rename(Path("/usr/X11R6"), Path("/usr/foobar"))
    assert original_fs - fs == {
        Path("/usr/X11R6"),
        Path("/usr/X11R6/bin"),
        Path("/usr/X11R6/include"),
        Path("/usr/X11R6/lib"),
        Path("/usr/X11R6/lib/tls"),
        Path("/usr/X11R6/man"),
        Path("/usr/X11R6/share"),
    }
    assert fs - original_fs == {
        Path("/usr/foobar"),
        Path("/usr/foobar/bin"),
        Path("/usr/foobar/include"),
        Path("/usr/foobar/lib"),
        Path("/usr/foobar/lib/tls"),
        Path("/usr/foobar/man"),
        Path("/usr/foobar/share"),
    }


if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



from pathlib import Path


def rm_tree(path: Path):  # https://stackoverflow.com/a/57892171/173003
    if not path.is_dir():
        return
    for child in path.iterdir():
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    path.rmdir()



class UnsupportedOSError(Exception):
    pass


class UnknownInodeError(ValueError):
    """
    The edited text contains an inode absent from the source text.
    """


class TabulationError(ValueError):
    """
    A new name contains a tabulation. Although such a character is valid on most
    platforms, in the edited text it probably results from a typo.
    """


class ValidationError(ValueError):
    """
    A new name includes invalid character(s) for a filename
    (depends on the target platform).
    """


class SeveralTargetsError(Exception):
    """
    Two distinct renaming targets are specified for the same source.
    """


class SeveralSourcesError(Exception):
    """
    Two distinct sources have the same renaming target,
    or a renaming target already exists and has no specified renaming.
    """


class DuplicatedClauseError(Exception):
    """
    A clause (source, target) is specified more than once.
    """


class RecoverableRenamingError(Exception):
    """
    Raised after a failed renaming, with the goal of triggering a rollback.
    """

class RetrieveDefaultsError(Exception):
    """
    Raised when the command for retrieving all defaults fail.
    """



import pytest

from pathlib import Path

import context
from src.secure_clauses import *


@pytest.fixture(scope="module")
def paths():
    return [Path(line) for line in Path("test/fhs.txt").read_text().strip().split("\n")]


@pytest.fixture()
def fs(paths):
    return FileSystem(paths)


def test_secure_clauses(fs):
    clauses = [
        (Path("/usr/X11R6/include"), "foobar"),
        (Path("/usr/local"), "bocal"),
        (Path("/usr/include"), "lib"),
        (Path("/usr/lib"), "include"),
    ]
    result = secure_clauses(fs, clauses)
    assert result == [
        (Path("/usr/X11R6/include"), Path("/usr/X11R6/foobar")),
        (Path("/usr/local"), Path("/usr/bocal")),
        (Path("/usr/include"), Path("/usr/NFXGG3DVMRSQ====-0")),
        (Path("/usr/lib"), Path("/usr/include")),
        (Path("/usr/NFXGG3DVMRSQ====-0"), Path("/usr/lib")),
    ]
    expected_fs_subset = {
        Path("/usr/bin"),
        Path("/usr/X11R6"),
        Path("/usr/libexec"),
        Path("/usr/X11R6/share"),
        Path("/usr/X11R6/foobar"),  # new path
        Path("/usr/src"),
        Path("/usr/X11R6/lib"),
        Path("/usr/bocal/share/man/man4"),  # new path
        Path("/usr/bocal/share/info"),  # new path
        Path("/usr/bocal/share/man/man2"),  # new path
        Path("/usr/bocal/src"),  # new path
        Path("/usr/bocal/libexec"),  # new path
        Path("/usr/bocal/share/man/man3"),  # new path
        Path("/usr/bocal/etc"),  # new path
        Path("/usr/bocal/bin"),  # new path
        Path("/usr/bocal/lib"),  # new path
        Path("/usr/bocal/include"),  # new path
        Path("/usr/bocal/share/man/man8"),  # new path
        Path("/usr/bocal"),  # new path
        Path("/usr/bocal/games"),  # new path
        Path("/usr/bocal/sbin"),  # new path
        Path("/usr/bocal/share/man/man6"),  # new path
        Path("/usr/bocal/share/man/man5"),  # new path
        Path("/usr/bocal/share/man"),  # new path
        Path("/usr/bocal/share/man/man1"),  # new path
        Path("/usr/bocal/share/man/man9"),  # new path
        Path("/usr/bocal/share/man/man7"),  # new path
        Path("/usr/bocal/share"),  # new path
        Path("/usr/bocal/share/man/mann"),  # new path
        Path("/usr/share"),
        Path("/usr/tmp"),
        Path("/usr/X11R6/man"),
        Path("/usr/include"),  # swapped path
        Path("/usr/lib"),  # swapped path
        Path("/usr/etc"),
        Path("/usr/sbin"),
        Path("/usr/games"),
        Path("/usr/X11R6/bin"),
    }
    assert expected_fs_subset.issubset(fs)


def test_secure_clauses_with_intermediate_clash(fs):
    clauses = [
        (Path("/usr/X11R6/lib"), "man"),
        (Path("/usr/X11R6/man"), "foo"),
        (Path("/usr/X11R6"), "bar"),
    ]
    result = secure_clauses(fs, clauses)
    assert result == [
        (Path("/usr/X11R6/lib"), Path("/usr/X11R6/NRUWE===-0")),
        (Path("/usr/X11R6/man"), Path("/usr/X11R6/foo")),
        (Path("/usr/X11R6/NRUWE===-0"), Path("/usr/X11R6/man")),
        (Path("/usr/X11R6"), Path("/usr/bar")),
    ]


def test_dict_of_clauses():
    clauses = [
        (Path("/foo/bar"), "buzz"),
        (Path("/ping"), "pong"),
        (Path("/ping"), "pong"),  # duplicates are forbidden
    ]
    with pytest.raises(DuplicatedClauseError) as offending_path:
        dict_of_clauses(clauses)
    assert offending_path.value.args[0] == "The clause '/ping' -> 'pong' is given twice."


def test_dict_of_clauses_when_not_a_function():
    clauses = [
        (Path("/foo/bar"), "buzz"),
        (Path("/ping"), "pong"),
        (Path("/ping"), "fail"),  # contradiction
    ]
    with pytest.raises(SeveralTargetsError) as offending_path:
        dict_of_clauses(clauses)
    assert offending_path.value.args[0] == "Two distinct renaming targets for '/ping': 'pong' and 'fail'."


def test_check_injectivity(fs):
    clauses = {
        Path("/usr/X11R6"): "R2D2",
        Path("/usr"): "User System Resources",
        Path("/usr/bin/X11"): "R2D2",
        Path("/usr/lib/X11"): "R2D2",
    }
    assert check_injectivity(fs, clauses) is None


def test_check_injectivity_with_internal_problem(fs):
    clauses = {
        Path("/usr/lib/games"): "foobar",
        Path("/usr/lib/X11"): "foobar",
    }
    with pytest.raises(SeveralSourcesError) as offending_path:
        check_injectivity(fs, clauses)
    assert offending_path.value.args[0] == "At least two distinct sources for '/usr/lib/foobar'."


def test_check_injectivity_with_external_problem(fs):
    clauses = {
        Path("/usr/lib/games"): "X11",
    }
    with pytest.raises(SeveralSourcesError) as offending_path:
        check_injectivity(fs, clauses)
    assert offending_path.value.args[0] == "At least two distinct sources for '/usr/lib/X11'."


def test_sorted_by_level():
    _ = "whatever"
    clause_dict = {
        Path("/sys"): _,
        Path("/tmp"): _,
        Path("/usr"): _,
        Path("/usr/X11R6"): _,
        Path("/usr/X11R6/bin"): _,
        Path("/usr/X11R6/include"): _,
        Path("/usr/X11R6/lib"): _,
        Path("/usr/X11R6/lib/tls"): _,
        Path("/usr/X11R6/man"): _,
        Path("/usr/X11R6/share"): _,
        Path("/usr/bin"): _,
        Path("/usr/bin/X11"): _,
    }
    expected = [
        (
            5,
            [
                (Path("/usr/X11R6/lib/tls"), _),
            ],
        ),
        (
            4,
            [
                (Path("/usr/X11R6/bin"), _),
                (Path("/usr/X11R6/include"), _),
                (Path("/usr/X11R6/lib"), _),
                (Path("/usr/X11R6/man"), _),
                (Path("/usr/X11R6/share"), _),
                (Path("/usr/bin/X11"), _),
            ],
        ),
        (
            3,
            [
                (Path("/usr/X11R6"), _),
                (Path("/usr/bin"), _),
            ],
        ),
        (
            2,
            [
                (Path("/sys"), _),
                (Path("/tmp"), _),
                (Path("/usr"), _),
            ],
        ),
    ]
    result = sorted_by_level(clause_dict)
    assert result == expected


if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



import pytest

import context
from src.get_editor_command import *


def test_get_editor_command_with_explicit_current_platform():
    platform = get_platform_long_string().partition("-")[0]
    command = get_editor_command(Path("foobar"), platform)
    assert command[0] in ["code", "subl", "open"]


def test_get_editor_command_with_detected_current_platform():
    command = get_editor_command(Path("foobar"))
    assert command[0] in ["code", "subl", "open"]


def test_get_editor_command_with_explicit_not_supported_platform():
    with pytest.raises(UnsupportedOSError):
        get_editor_command(Path("foobar"), "Windows")


def test_get_editor_command_with_explicit_wrong_platform():
    platform = get_platform_long_string().partition("-")[0]
    with pytest.raises(Exception):
        get_editor_command(Path("foobar"), "macOS" if platform == "Linux" else "Linux")


if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



import subprocess
import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List

sys.path[0:0] = ["."]

from src.file_system import FileSystem
from src.get_editable_text import get_editable_text
from src.get_editor_command import get_editor_command
from src.logger import logger
from src.parse_edited_text import parse_edited_text
from src.paths_to_inodes_paths import paths_to_inodes_paths
from src.printer import print_
from src.renamings import Renamer
from src.secure_clauses import secure_clauses
from src.user_errors import *
from src.user_types import EditedText


def main_wrapper():
    previous_log_text = logger.get_contents()
    logger.create_new_log_file()
    logger.info("Starting the program.")
    main(previous_log_text)
    logger.info("Exiting the program.")


def main(previous_log_text: str):
    logger.info("Parsing arguments.")
    args = cli_arguments()
    logger.info("Parsing arguments done.")

    if not args.undo:
        logger.info("Constructing the list of items to rename.")
        paths: List[Path] = []
        if args.paths:
            logger.info("A list of items to rename was provided.")
            paths.extend(map(Path, args.paths))
        if args.file:
            logger.info("A file containing the paths to rename was provided.")
            paths.extend(Path(path) for path in Path(args.file).read_text().split("\n") if path)
        if not paths:
            logger.info("No paths to rename were provided.")
            return print_.abort("Please provide at least one file to rename.")
        logger.info("Running on path list.")
        run_on_path_list(paths)
        logger.info("Running on path list done.")
    else:
        logger.info("Undoing renamings.")
        renamer = Renamer()
        try:
            arcs_for_undoing = renamer.get_arcs_for_undoing(previous_log_text)
            logger.info(str(arcs_for_undoing))
            message = renamer.perform_renamings(arcs_for_undoing)
            print_.success(
                f"The previous renaming session was undone. {message} "
                "Launch Suprenam again to restore."
            )
        except RecoverableRenamingError:
            try:
                renamer.rollback_renamings()
                print_.abort("The renamings failed, but were successfully rolled back.")
            except Exception as e:
                return print_.fail(f"Unrecoverable failure during rollback: {e}")
        except Exception as e:  # Unknown problem with the log file, e.g. not found
            return print_.fail(f"Unrecoverable failure during undo: {e}")
        logger.info("Undoing renamings done.")


def run_on_path_list(paths: List[Path]):

    logger.info("Creating a mapping from inodes to paths.")
    try:
        inodes_paths = paths_to_inodes_paths(paths)
        logger.info("Creating a mapping from inodes to paths done.")
    except Exception as e:
        return print_.abort(str(e))

    logger.info("Creating a temporary text file for the list to be edited.")
    try:
        editable_file_path = Path(NamedTemporaryFile(mode="w+", delete=False, suffix=".tsv").name)
        logger.info(f"Editable file path: {repr(editable_file_path)}.")
    except Exception as e:
        return print_.abort(f"Failed to create a temporary file: {e}")

    logger.info("Populating the temporary text file with the list to be edited.")
    try:
        editable_file_path.write_text(get_editable_text(inodes_paths))
        logger.info(f"Editable file content: populated.")
    except Exception as e:
        return print_.abort(f"Failed to populate the temporary file: {e}")

    logger.info("Retrieving a command to edit the temporary text file.")
    try:
        editor_command = get_editor_command(editable_file_path)
        logger.info(f"The command is {' '.join(editor_command)}.")
    except UnsupportedOSError:
        return

    logger.info("Opening the editable text file in the editor and waiting it to be closed.")
    try:
        subprocess.run(editor_command, check=True)
        logger.info("Command executed without process error.")
    except subprocess.CalledProcessError:
        return print_.abort(f"The command {' '.join(editor_command)} failed.")

    logger.info("Retrieving the content of the edited text file.")
    try:
        edited_text = EditedText(editable_file_path.read_text())
        logger.info("Line count in the edited text file: %s." % edited_text.count("\n"))
    except Exception as e:
        return print_.abort(f"Failed to read the edited text file: {e}")

    logger.info("Parsing the edited text into renaming clauses.")
    try:
        clauses = parse_edited_text(edited_text, inodes_paths)
        logger.info(f"Parsed edited text into {len(clauses)} clauses.")
    except Exception as e:
        return print_.abort(str(e))

    logger.info("Converting the clauses into a 'safe' sequence of renamings.")
    try:
        arcs = secure_clauses(FileSystem(), clauses)
        logger.info(f"Converted clauses into {len(arcs)} arcs.")
    except Exception as e:
        return print_.abort(str(e))

    logger.info("Performing the actual renamings.")
    renamer = Renamer()
    try:
        message = renamer.perform_renamings(arcs)
        return print_.success(message)
    except RecoverableRenamingError:
        logger.warning("Renaming performed with a recoverable error.")
        try:
            message = renamer.rollback_renamings()
            return print_.abort(f"The renamings failed, but don't worry: {message}.")
        except Exception as e:
            return print_.fail(
                f"Unrecoverable failure during rollback: {e}"
                "Some files may have been renamed, some not."
                "Please check the log file at `~/.suprenam/log.txt`."
            )


def cli_arguments():
    """
    CLI argument parser.

    Returns:
        `parser.parse_args()` dict containing the parsed arguments.
    """
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        usage=f"\n{Path(__file__).name} [-p paths] [-f file] [-h help]",
        description=f"\nFILE RENAMER",
    )

    parser.add_argument(
        "-p",
        "--paths",
        nargs="+",
        help=f"The paths to rename.",
        action="store",
    )

    parser.add_argument(
        "-f",
        "--file",
        help=f"Parse paths stored in a file (newline separated).",
        action="store",
    )

    parser.add_argument(
        "-u",
        "--undo",
        help=f"Undo completed renamings from the previous session.",
        action="store_true",
    )

    return parser.parse_args()


if __name__ == "__main__":  # pragma: no cover
    main_wrapper()



import re
import subprocess
from pathlib import Path
from platform import platform as get_platform_long_string
from shutil import which
from typing import Optional

from src.user_errors import *


OS = {
    "macOS": {
        "QUERY_ALL_DEFAULTS_COMMAND": "defaults read com.apple.LaunchServices/com.apple.launchservices.secure LSHandlers",
        "EXTRACT_EDITOR": r'(?ms)\s*\{\s*LSHandlerContentType = "public\.plain-text";\s*LSHandlerPreferredVersions =\s*\{\s*LSHandlerRoleAll = "-";\s*\};\s*LSHandlerRoleAll = "([\w.]+)";',
        "CUSTOM_EDITOR_COMMAND": {
            "com.microsoft.vscode": ["code", "-w"],
            "com.sublimetext.3": ["subl", "-w"],
        },  # TODO: add different versions of Sublime Text
        "FALLBACK_EDITOR_COMMAND": [
            "open",
            "-n",  # open a new instance of the application even if one is already running
            "-e",  # open with TextEdit
            "-W",  # block until the **application** is closed (even if it was already running).
            #        This is far from ideal, but there is currently no per-window way to check
        ],
    },
    "Linux": {
        "QUERY_ALL_DEFAULTS_COMMAND": "xdg-mime query default text/plain",
        "EXTRACT_EDITOR": r"^(.*)\.desktop$",
        "CUSTOM_EDITOR_COMMAND": {
            "code": [
                "code",
                "-w",  # wait for the file to be closed before returning
            ],
            "sublime_text": ["subl", "-w"],
        },
        "FALLBACK_EDITOR_COMMAND": ["open", "-w"],
    },
    # TODO: add Windows
}


def is_tool(name):  # https://stackoverflow.com/a/34177358/173003
    """Check whether `name` is on PATH and marked as executable."""
    return which(name) is not None


def get_editor_command(path: Path, platform: Optional[str] = None) -> list:
    """
    Retrieve a command launching a text editor on a given text file.
    Args:
        path: the path to the text file to edit.
    Returns:
        A list of strings representing the command to launch the system's default text editor
        on the given text file. If no default text editor is defined, a suitable fallback command
        is returned.
    Raises:
        UnsupportedOSError: if the OS dictionary defines no key for the given operating system name.
    """
    platform = platform or get_platform_long_string().partition("-")[0]
    os_dict = OS.get(platform)
    if not os_dict:
        raise UnsupportedOSError(
            f"Unsupported operating system: {platform}. "
            f"Supported operating systems are: {', '.join(OS.keys())}."
        )

    try:
        output = subprocess.run(
            str(os_dict["QUERY_ALL_DEFAULTS_COMMAND"]).split(),  # make mypy happy
            encoding="utf-8",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        ).stdout
    except Exception as e:
        raise RetrieveDefaultsError(f"Failed to retrieve the defaults: {e}.")

    match = re.search(str(os_dict["EXTRACT_EDITOR"]), output)  # make mypy happy
    custom_editor_handler = match.group(1) if match else ""

    # Check whether the user has defined a custom editor
    command = os_dict["CUSTOM_EDITOR_COMMAND"].get(custom_editor_handler)  # type: ignore
    if command and is_tool(command[0]): # is the command both known and installed on the system?
        return command + [str(path)]

    # Otherwise, try to find another editor which is both known and installed on the system
    for command in os_dict["CUSTOM_EDITOR_COMMAND"].values():  # type: ignore
        if is_tool(command[0]):
            return command + [str(path)]

    # Otherwise, return the fallback command defined by the system
    # (on macOS, this is TextEdit, which lacks an option to wait for the file to be closed)
    return list(os_dict["FALLBACK_EDITOR_COMMAND"]) + [str(path)]  # make mypy happy



from pathlib import Path
from typing import Dict, NamedTuple, NewType

Name = NewType("Name", str)


class Clause(NamedTuple):
    path: Path
    new_name: Name


ClauseMap = Dict[Path, Name]


class Arc(NamedTuple):
    source: Path
    target: Path


Inode = NewType("Inode", int)

InodesPaths = Dict[Inode, Path]

EditableText = NewType("EditableText", str)
EditedText = NewType("EditedText", str)



import pytest

from pathlib import Path

import context
import src.secure_clauses as sc
from src.file_system import FileSystem
from extract_examples import extract_examples
from reformat_examples import main as reformat_examples

EXCEPTIONS = {
    "FileNotFoundError": __builtins__["FileNotFoundError"],
    "SeveralTargetsError": getattr(sc, "SeveralTargetsError"),
    "SeveralSourcesError": getattr(sc, "SeveralSourcesError"),
    "DuplicatedClauseError": getattr(sc, "DuplicatedClauseError"),
}

PATHS = [Path(line) for line in Path("test/fhs.txt").read_text().strip().split("\n")]
EXAMPLES_MD_PATH = Path("test/examples.md")

reformat_examples(EXAMPLES_MD_PATH)


@pytest.mark.parametrize(
    "i, title, example, expected",
    [d.values() for d in extract_examples(EXAMPLES_MD_PATH)[:]],
)
def test(i, title, example, expected):
    clauses = [(Path(row[0]), row[1]) for row in example]
    fs = FileSystem(PATHS)
    original_fs = set(fs)
    if isinstance(expected, tuple):
        (exception_name, expected_culprit) = expected
        with pytest.raises(EXCEPTIONS[exception_name]) as culprit:
            sc.secure_clauses(fs, clauses)
        assert culprit.value.args[0] == expected_culprit
        assert fs == original_fs
    else:
        safe_clauses = sc.secure_clauses(fs, clauses)
        additions = {Path(row[1]) for row in expected}
        deletions = {Path(row[0]) for row in expected}
        (additions, deletions) = (additions - deletions, deletions - additions)
        assert original_fs - fs == deletions
        assert fs - original_fs == additions


if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



import sys
from pathlib import Path
from src.logger import logger


class ColorPrinter:
    """Print colorized messages on the standard output."""

    # Optional ANSI color codes
    OK = "\033[92m"
    WARNING = "\033[1m\033[38;5;166m"
    FAIL = "\033[1m\033[91m"
    RESET = "\033[0m"

    def __call__(self, message: str):
        print(message)

    def newline(self):
        print()

    def success(self, message: str):
        logger.info(f"Success: {message}")
        print(f"{self.OK}Success: {message}{self.RESET}", file=sys.stderr)

    def abort(self, message: str):
        logger.warning(f"Abort: {message}")
        print(f"{self.WARNING}No renamings were performed: {message}{self.RESET}", file=sys.stderr)

    def fail(self, message: str):
        logger.error(f"Fail: {message}")
        print(f"{self.FAIL}Fatal error: {message}{self.RESET}", file=sys.stderr)


class PlatypusPrinter:
    """Print uncolorized messages on the standard output and Platypus' alert."""

    def __init__(self):
        self.buffer = []

    def __call__(self, message: str):
        print(message)

    def newline(self):
        print("\r")

    def success(self, message: str):
        logger.info(f"Success: {message}")
        print(f"ALERT:Success|{message}")

    def abort(self, message: str):
        logger.warning(f"Abort: {message}")
        print(f"ALERT:No renamings were performed|{message}")

    def fail(self, message: str):
        logger.error(f"Fail: {message}")
        print(f"ALERT:Fatal error|{message}")


print_ = ColorPrinter()
if Path.cwd().name == "Resources":
    # We are in the app bundle (macOS)
    print_ = PlatypusPrinter()  # type: ignore



from pathlib import Path
from pathvalidate import ValidationError

import context
from src.get_editable_text import *

import pytest
import re

data = [  # NB: in the triple-quoted strings, inodes and names are tab-separated.
    (
        "Nothing -> empty result",
        {},
        "",
    ),
    (
        "Empty path -> empty string",
        {
            2612647906: Path(),
        },
        "",
    ),
    (
        "Root -> empty string",
        {
            2612647906: Path("/"),
        },
        "",
    ),
    (
        "A path containing two consecutive '/'",
        {
            2612647906: Path("/foobar//"),
        },
        """
        2612647906	foobar
        """,
    ),
    (
        "A relative path in the current working directory",
        {
            2612647906: Path("foobar"),
        },
        """
        2612647906	foobar
        """,
    ),
    (
        "A path ending with a '/'",
        {
            2612647906: Path("foobar/"),
        },
        """
        2612647906	foobar
        """,
    ),
    (
        "A bunch of siblings -> (unique) parent not displayed",
        {
            2612647906: Path("/usr/lib/gcc-lib"),
            3263123972: Path("/usr/lib/locale"),
            549534358: Path("/usr/lib/X11"),
            1582956841: Path("/usr/lib/tls"),
            1981263235: Path("/usr/lib/games"),
        },
        """
            1981263235	games
            2612647906	gcc-lib
            3263123972	locale
            1582956841	tls
            0549534358	X11
        """,
    ),
    (
        "An arbitrary arborescence -> siblings grouped by parents and sorted",
        {
            1421832123: Path("/usr/bin/X11"),
            2612647906: Path("/usr/lib/gcc-lib"),
            2784148792: Path("/usr/lib"),
            3263123972: Path("/usr/lib/locale"),
            1482466386: Path("/usr/bin"),
            549534358: Path("/usr/lib/X11"),
            1582956841: Path("/usr/lib/tls"),
            1981263235: Path("/usr/lib/games"),
        },
        """
            /usr
            1482466386	bin
            2784148792	lib
    
            /usr/bin
            1421832123	X11
    
            /usr/lib
            1981263235	games
            2612647906	gcc-lib
            3263123972	locale
            1582956841	tls
            0549534358	X11
        """,
    ),
    (
        "A bunch of siblings （shown in default order) -> names sorted in natural order",
        {
            1421832123: Path("/usr/lib/Apple"),
            2612647906: Path("/usr/lib/apple14,689"),
            2784148792: Path("/usr/lib/apple15"),
            1482466380: Path("/usr/lib/Banana_b"),
            3263123972: Path("/usr/lib/banana_a"),
            1482466386: Path("/usr/lib/banana_c"),
            549534358: Path("/usr/lib/version-1.10"),
            1582956841: Path("/usr/lib/version-1.11"),
            1981263235: Path("/usr/lib/version-1.9"),
            1981263299: Path("/usr/lib/version-2.0"),
        },
        """
            1421832123	Apple
            2612647906	apple14,689
            2784148792	apple15
            3263123972	banana_a
            1482466380	Banana_b
            1482466386	banana_c
            1981263235	version-1.9
            0549534358	version-1.10
            1582956841	version-1.11
            1981263299	version-2.0
        """,
    ),
    (
        "The parents are also sorted in natural order",
        {
            1482466380: Path("/usr/Banana_b/foobar"),
            3263123972: Path("/usr/banana_a/foobar"),
            1482466386: Path("/usr/banana_c/foobar"),
        },
        """
        /usr/banana_a
        3263123972	foobar

        /usr/Banana_b
        1482466380	foobar

        /usr/banana_c
        1482466386	foobar
        """,
    ),
]


@pytest.mark.parametrize("title, inodes_paths, expected", data)
def test_get_editable_text(title, inodes_paths, expected):
    print(title)
    expected = expected.strip().replace("    ", "").split("\n")
    result = get_editable_text(inodes_paths).strip().split("\n")
    print("\n".join(result))
    assert len(expected) == len(result)
    for (expected_line, result_line) in zip(expected, result):
        assert expected_line == result_line




if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



import logging
from pathlib import Path
import os, datetime, shutil


class Logger:

    LOG_DIR = Path.home() / ".suprenam"
    LOG_NAME = "log.txt"
    LOGS_TO_KEEP = 10

    def __init__(self, path: Path = LOG_DIR / LOG_NAME):
        """Ensure the log directory exists."""
        path.parent.mkdir(
            parents=True,  # any missing parents of this path are created as needed
            exist_ok=True,  # if the directory already exists, do not raise an exception
        )
        self.path = path

    def create_new_log_file(self):
        """Remove all handlers associated with the root logger object and create a NEW log file."""
        self.backup_current_log_file()
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
        logging.basicConfig(filename=self.path, filemode="w", level=logging.DEBUG)

    def backup_current_log_file(self):
        """
        Copy the current log file with a timestamp appended.
        Clean up the directory by keeping only the last copies.
        """
        if self.LOGS_TO_KEEP == 0:
            return
        try:
            modified = os.path.getmtime(self.path)
        except FileNotFoundError:
            return
        suffix = datetime.datetime.fromtimestamp(modified).strftime("%Y-%m-%d_%H-%M-%S.%f")
        destination = f"{self.path.parent}/log_{suffix}.txt"
        shutil.copy2(self.path, destination)
        for file in sorted(self.LOG_DIR.glob("log_*.txt"), reverse=True)[self.LOGS_TO_KEEP :]:
            file.unlink()

    def get_contents(self):  # pragma: no cover
        if self.path.is_file():
            return self.path.read_text().strip()
        else:
            return ""

    def erase_contents(self):
        self.path.write_text("")

    def warning(self, *args, **kwargs):
        logging.warning(*args, **kwargs)

    def info(self, *args, **kwargs):
        logging.info(*args, **kwargs)

    def error(self, *args, **kwargs):
        logging.error(*args, **kwargs)


logger = Logger()



import os
import setuptools
from pathlib import Path
from time import time

setuptools.setup(
    name="suprenam",
    version=os.environ.get("RELEASE_VERSION") if os.environ.get("RELEASE_VERSION") != "main" else f"0.9.9-beta.{int(time())}",
    author="Aristide Grange & Alexandre Perlmutter",
    author_email="alexandre.perlmutter@gmail.com",
    description="Easily rename files and folders via your favorite text editor.",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/poponealex/suprenam",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(include=['src', 'src.*']),
    python_requires=">=3.6",
    entry_points={"console_scripts": ["suprenam=src.suprenam:main"]},
)



from pathlib import Path
from pathvalidate import ValidationError

import context
from src.parse_edited_text import *

import pytest
import re


data = [
    (
        "Parse a file which doesn't end with a newline.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	foobar
            456	club""",
        [
            Clause(Path("/home/foo"), "foobar"),
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which is missing an inode.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            	foobar
            456	club
        """,
        [
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which contains a name starting with a space.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	 foobar
            456	club
        """,
        [
            Clause(Path("/home/foo"), " foobar"),
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which contains a name made of multiple whitespaces.",
        ["macOS", "Linux"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	    
            456	club
        """,
        [
            Clause(Path("/home/foo"), "    "),
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which contains an inode preceeded by a whitespace.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
             123	foobar
            456	club
        """,
        [
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which contains a name containing a '\\' character on macOS/Linux.",
        ["macOS", "Linux"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	foo\\bar
            456	club
        """,
        [
            Clause(Path("/home/foo"), "foo\\bar"),
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which contains a name containing a '\\' character on Windows.",
        ["Windows"],
        ValidationError,
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	foo\\bar
            456	club
        """,
        [],
    ),
    (
        "Parse a file which contains a name that is only digits.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	12345
            456	club
        """,
        [
            Clause(Path("/home/foo"), "12345"),
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which contains too many inodes (superset).",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	foobar
            456	club
            789 spam
        """,
        [
            Clause(Path("/home/foo"), "foobar"),
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which doesn't contain all the inodes (subset).",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	foobar
        """,
        [
            Clause(Path("/home/foo"), "foobar"),
        ],
    ),
    (
        "Parse a file which contains only one modification.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
            789: Path("/home/spam"),
        },
        """
            123	foo
            456	club
            789	spam
        """,
        [
            Clause(Path("/home/bar"), "club"),
        ],
    ),
    (
        "Parse a file which doesn't contain any modification.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        """
            123	foo
            456	bar
        """,
        [],
    ),
    (
        "Parse a file which is empty.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
        },
        "",
        [],
    ),
    (
        "Parse a file which contains paths with different parents.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
            456: Path("/home/bar"),
            789: Path("/etc/lib"),
            999: Path("/hello/world"),
        },
        """
            /home
            123	foobar
            456	club

            /etc
            789	book

            /hello
            999	moon
        """,
        [
            Clause(Path("/home/foo"), "foobar"),
            Clause(Path("/home/bar"), "club"),
            Clause(Path("/etc/lib"), "book"),
            Clause(Path("/hello/world"), "moon"),
        ],
    ),
    (
        "Parse a file which is missing a name.",
        ["universal"],
        ValidationError,
        {
            123: Path("/home/foo"),
        },
        """
            123	
        """,
        [],
    ),
    (
        "Parse a file which contains an illegal '/' character.",
        ["universal"],
        ValidationError,
        {
            123: Path("/home/foo"),
        },
        """
            123	foo/bar
        """,
        [],
    ),
    (
        "Parse a file whose name refers to the parent directory.",
        ["universal"],
        ValidationError,
        {
            123: Path("/home/foo"),
        },
        """
            123	../foo
        """,
        [],
    ),
    (
        "Parse a file which contains a non-existent inode.",
        ["universal"],
        UnknownInodeError,
        {
            123: Path("/home/foo"),
        },
        """
            222	foobar
        """,
        [],
    ),
    (
        "Parse a file which contains multiple tabs in between the inode and the new name.",
        ["universal"],
        TabulationError,
        {
            123: Path("/home/foo"),
        },
        """
            123				foobar
        """,
        [],
    ),
    (
        "Parse a file which contains several times the same inode with different targets.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
        },
        """
            123	foobar
            123	barfoo
        """,
        [ # This case will be filtered out by secure_clauses().
            Clause(Path("/home/foo"), "foobar"),
            Clause(Path("/home/foo"), "barfoo"),
        ],
    ),
    (
        "Parse a file which contains several times the same inode with the same target.",
        ["universal"],
        "no_exception",
        {
            123: Path("/home/foo"),
        },
        """
            123	foobar
            123	foobar
        """,
        [ # This case will be filtered out by secure_clauses().
            Clause(Path("/home/foo"), "foobar"),
            Clause(Path("/home/foo"), "foobar"),
        ],
    ),
    (
        "Parse a file which contains some numerical parents.",
        ["universal"],
        "no_exception",
        {
            123: Path("321/foo"),
            456: Path("321/bar"),
            789: Path("654/lib"),
            999: Path("987/world"),
        },
        """
            321
            123	foobar
            456	club

            654
            789	book

            987
            999	moon
        """,
        [
            Clause(Path("321/foo"), "foobar"),
            Clause(Path("321/bar"), "club"),
            Clause(Path("654/lib"), "book"),
            Clause(Path("987/world"), "moon"),
        ],
    ),
]


split_startwith_tabs = re.compile(r"(?m)^([ ]{4})*(.*)$").findall


@pytest.mark.parametrize("title, platforms, exception, inodes_paths, text, expected", data)
def test_parse_edited_text(title, platforms, exception, inodes_paths, text, expected):
    print(title)
    text = "\n".join(line for (_, line) in split_startwith_tabs(text))
    for platform in platforms:
        actual = lambda: parse_edited_text(text, inodes_paths, platform)
        if exception != "no_exception":
            with pytest.raises(exception):
                actual()
        else:
            print(actual())
            print(expected)
            assert actual() == expected


if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



from typing import List, Optional

import pathvalidate

from src.printer import print_
from src.user_errors import *
from src.user_types import Clause, EditedText, Inode, InodesPaths, Name


def parse_edited_text(
    text: EditedText,
    inodes_paths: InodesPaths,
    platform: Optional[str] = None,
) -> List[Clause]:
    """
    Parse the text in which the user has modified some filenames to specify a renaming.

    Args:
        text: the text with the effective renamings.
        inodes_paths: dict containing each path (value) with its inode (key).
        platform: OS on which the renamings will be performed. Without a value `path_validate` will
            detect the platform automatically.

    Raises:
        UnknownInodeError: the edited text contains an inode absent from the source text.
        TabulationError: a new name contains a tabulation. Although such a character is valid on most
            platforms, in the edited text it probably results from a typo.
        pathvalidate.ValidationError: a new name includes invalid character(s) for a filename
            (depends on the target platform).

    Returns:
        A list of clauses (path, new_name).
    """
    result = []
    for line in text.split("\n"):
        head, *tail = line.split("\t", maxsplit=1)
        if not tail:
            continue
        if not head.isdigit():
            continue
        inode = Inode(int(head))

        path = inodes_paths.get(inode)
        if path is None:
            raise UnknownInodeError(f"Unknown inode {inode}.")

        new_name = tail[0]
        if "\t" in new_name:
            raise TabulationError(f"Illegal tabulation in the new name: '{new_name}'.")
        if path.name == new_name:
            continue
        try:
            pathvalidate.validate_filename(new_name, platform=platform or "auto")
        except pathvalidate.ValidationError:
            raise ValidationError(f"Invalid character(s) in the new name: '{new_name}'.")
        result.append(Clause(path, Name(new_name)))
    return result



from itertools import groupby
from typing import Iterable, List, Tuple

from src.file_system import FileSystem
from src.user_types import Clause, ClauseMap, Name, Arc
from src.user_errors import DuplicatedClauseError, SeveralTargetsError, SeveralSourcesError
from src.printer import print_


def secure_clauses(file_system: FileSystem, clauses: List[Clause]) -> List[Arc]:
    """Construct a "safe" version of the given renaming clauses and update the file system.

    The resulting sequence is a reordered copy of the given clauses, with potentially the
    insertion of intermediate renamings, in such a way that they can be executed sequentially in
    the concrete file system without any collision or other adverse effect.

    Args:
        clauses (List[Clause]): A collection of renaming clauses of the form `(path, new_name)`
            in no particular order.
        file_system (FileSystem): A file system constructed either with an empty or nonempty list
            of paths:
            - empty: the arborescence will be populated by calling `update_with_source_paths()`
                with a list of source paths: this method will add not only the given paths, but
                their actual siblings too.
            - nonempty: used with a prepopulated test file system, like `test/fhs.txt`.

    Raises:
        SeveralTargetsError: when two distinct renaming targets are specified for the same source.
        SeveralSourcesError: when two distinct sources have the same renaming target, or when a
            renaming target already exists and has no specified renaming.
        FileNotFound: when a source does not exist in the file system.

    Side Effect:
        The underlying (pure) file system is brought to the expected final state (i.e., as if
        all the renamings would have been executed).

    Returns:
        List[Arc]: A "safe" version of the given renaming clauses, presented as a list of
            source and target paths.
    """
    clause_dict = dict_of_clauses(clauses)
    try:
        file_system.update_with_source_paths(clause_dict.keys())
    except FileNotFoundError as e:
        raise FileNotFoundError(f"File not found: '{e}'.")
    check_injectivity(file_system, clause_dict)
    clauses_by_levels = sorted_by_level(clause_dict)
    safe_clauses = []
    for (level, clauses) in clauses_by_levels:
        i = 0
        while i < len(clauses):  # `clauses` sequence may grow
            (path, new_name) = clauses[i]
            new_path = path.with_name(new_name)
            if new_path in file_system:
                new_path = file_system.non_existing_sibling(path)
                clauses[i] = Clause(path, Name(new_path.name))
                clauses.append(Clause(new_path, new_name))
            file_system.rename(path, new_path)
            i += 1
        safe_clauses.extend(clauses)
    return [Arc(path, path.parent / new_name) for (path, new_name) in safe_clauses]


def dict_of_clauses(clauses: Iterable[Clause]) -> ClauseMap:
    """Make a dictionary from the given clauses.

    During its construction, raise an error iff there exists at least one couple of clauses
    `(path, new_name_1)` and `(path, new_name_2)` (SeveralTargetsError if the new names are
    distinct, and DuplicatedClauseError otherwise).

    Args:
        clauses (List[Clause]): A list of couples of the form `(path, new_name)`.

    Raises:
        SeveralTargetsError: Two distinct new names are specified for the same path.
        DuplicatedClauseError: The same name is specified as least twice for the same path.

    Returns:
        ClauseMap: A dictionary associating paths to new names.
    """
    result: ClauseMap = {}
    for (path, new_name) in clauses:
        if path in result:
            if result[path] == new_name:
                raise DuplicatedClauseError(f"The clause '{path}' -> '{new_name}' is given twice.")
            else:
                raise SeveralTargetsError(
                    f"Two distinct renaming targets for '{path}': "
                    f"'{result[path]}' and '{new_name}'."
                )
        result[path] = new_name
    return result


def check_injectivity(file_system: FileSystem, clauses: ClauseMap):
    """Check that no resulting path has more than one antecedent.

    This function detects the following problems:

    1. There are two sibling nodes being renamed onto the same target path.
    2. There is a node being renamed onto a target path which already exists in the
       file system without being itself renamed (i.e., not the source of any clause).

    Args:
        clauses (ClauseMap): A dictionary associating paths to new names.

    Raises:
        SeveralSourcesError: A resulting path has two distinct antecedents.
    """
    already_seen = set()
    for (path, new_name) in clauses.items():
        new_path = path.with_name(new_name)
        if new_path in already_seen or (new_path in file_system and new_path not in clauses):
            raise SeveralSourcesError(f"At least two distinct sources for '{new_path}'.")
        already_seen.add(new_path)


def sorted_by_level(clauses: ClauseMap) -> List[Tuple[int, List[Clause]]]:
    """Order and group the items of a clause dictionary with the most nested first."""
    items = (Clause(*item) for item in clauses.items())  # required by mypy (as of v. 0.812)
    sorted_clauses = sorted(items, key=lambda clause: len(clause.path.parts), reverse=True)
    grouped_clauses = groupby(sorted_clauses, key=lambda clause: len(clause.path.parts))
    return [(level, list(clauses)) for (level, clauses) in grouped_clauses]



import re
from pathlib import Path
from typing import List

from src.printer import print_
from src.user_errors import RecoverableRenamingError
from src.user_types import Arc
from src.logger import logger


class Renamer:
    def __init__(self, testing: bool = False):
        if testing:
            logger.create_new_log_file()

    def perform_renamings(self, arcs: List[Arc]):
        """
        Perform the renamings specified by the list of arcs (source path, target path), and
        log them one by one.

        Args:
            arcs: list of couples (source_path, target_path)
        """
        n = len(arcs)
        print_(f"Renaming {n} items...")
        logger.info(f"{n} item{'s'[:n^1]} to rename.")
        try:
            self.rename_and_log_all_files(arcs)
            logger.info(f"{n} item{'s'[:n^1]} renamed.")
            print_(f"{n} item{'s'[:n^1]} renamed.")
            if n == 0:
                return f"Nothing to rename."
            elif n == 1:
                return f"1 item was renamed."
            else:
                return f"All {n} items were renamed."
        except Exception as e:
            logger.warning(f"perform_renamings: {e}")
            raise RecoverableRenamingError(f"{e}.")

    def rollback_renamings(self):
        """
        Rollback the first renaming operations.

        The inverse renaming are appended to the log file.
        """
        n = len(self.arcs_to_rollback)
        print_(f"Rolling back the first {n} renaming{'s'[:n^1]}...")
        logger.info(f"{n} renaming{'s'[:n^1]} to roll back.")
        try:
            self.rename_and_log_all_files(self.arcs_to_rollback)
            logger.info(f"{n} renaming{'s'[:n^1]} rolled back.")
            if n == 0:
                return f"there was nothing to roll back."
            elif n == 1:
                return f"the only renaming was rolled back."
            else:
                return f"all {n} renamings were rolled back."
        except Exception as e:
            logger.error(f"rollback_renamings: {e}")
            raise

    def get_arcs_for_undoing(
        self,
        previous_log_text: str,
        get_logged_arcs=re.compile(r"(?m)^\w+:\w+:SOURCE:(.+)\tTARGET:(.+)").findall,
    ):
        """Read a log file and calculate the reversed renamings."""
        if re.search(r"(?m)^ERROR:", previous_log_text):  # The log file contains an error.
            raise ValueError("The previous rollback failed. Undoing is not possible.")
        arcs = []
        for (source, target) in reversed(get_logged_arcs(previous_log_text)):
            arcs.append(Arc(Path(target), Path(source)))
        return arcs

    def rename_and_log_all_files(self, arcs: List[Arc]):
        self.arcs_to_rollback: List[Arc] = []
        for (source, target) in arcs:
            source.rename(target)
            self.arcs_to_rollback.insert(0, Arc(target, source))
            logger.info(f"SOURCE:{source}\tTARGET:{target}")
        self.print_arcs(arcs)
        print_.newline()

    def print_arcs(self, arcs: List[Arc]):
        previous_parent = Path()
        for (source, target) in arcs:
            if source.parent != previous_parent:
                print_.newline()
                print_(f"{source.parent}")
                previous_parent = source.parent
            print_(f"{source.name} -> {target.name}")
